/*
	Immaavs plugin for AviSynth -- ImageMagick Reader/Writer for AviSynth.

	Copyright (C) 2010 Wilbert Dijkhof
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at:
	Wilbert Dijkhof
	wilbertdijkhof at yahoo.com
*/

#include "stdafx.h"
#include "avisynth.h"
#include <windows.h>
#include <magick/MagickCore.h>
#include <sstream>
#include <iomanip>   // setfill, setw
#include <cassert>   // assert
#include "utility.h" // our file and directory path functions
#include "text-overlay.h"  // for messages
#include "filesearch.h"  // index images
using namespace std;

// http://magick.imagemagick.org/script/magick-core.php
// http://htdig.mirror.directi.com/manual/en/function.imagick-getimageheight.php

// immamog: http://forum.gleitz.info/showthread.php?t=33835&page=11 (using special effects)


class ImmaRead : public IClip {
/**
  * Class to read image sequences into video buffers
 **/
public:
	ImmaRead(const char filename[], const int _start, const int _end, const int _force_size_x, const int _force_size_y,
				bool _animation, bool _gapless, int _density, bool _info, const char* _magick_home, IScriptEnvironment* env);
	~ImmaRead();
	
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env); 
	
	void __stdcall GetAudio(void* buf, __int64 start, __int64 count, IScriptEnvironment* env) {}
	const VideoInfo& __stdcall GetVideoInfo() { return vi; }
	bool __stdcall GetParity(int n) { return false; }
	void __stdcall SetCacheHints(int cachehints, int frame_range) { };

private:
	const int start, end, force_size_x, force_size_y;
	bool animation, gapless, info;
	const char* magick_home;
	char bufsize[30], fsize_x[4], fsize_y[4], dens[10];
	bool image_seq, force_noresize;
	VideoInfo vi;
	string fullname;   // full pathname supplied to the ctor
	string old_path;   // working directory (location of script); *may change* during runtime
	string img_path;   // new directory (location of image)
	string img_name;   // image filename (without path)
	FileSearch fs;     // contains the found images
	char message[255];
	PClip message_clip;
};

ImmaRead::ImmaRead(const char filename[], const int _start, const int _end, const int _force_size_x,
				   const int _force_size_y, bool _animation, bool _gapless, int _density, bool _info,
				   const char* _magick_home, IScriptEnvironment* env) :
	fullname(filename), old_path(), img_path(), img_name(), start(_start), end(_end), force_size_x(_force_size_x),
	force_size_y(_force_size_y), animation(_animation), gapless(_gapless), info(_info), magick_home(_magick_home) {

	size_t ext_pos = fullname.rfind('.');
	string extension = (ext_pos != string::npos) ? fullname.substr(ext_pos + 1) : "";
	const char* ext = extension.c_str();
	char* pPath;
	char img_name2[MaxTextExtent];

	// raw camera formats: CR2 DCR DNG ERF MEF MRW NEF ORF PEF RAF RAW SR2 X3F 3FR
	if (!(!lstrcmpi(ext, "jpeg") || !lstrcmpi(ext, "jpe") || !lstrcmpi(ext, "jpg") || !lstrcmpi(ext, "jp2") || 
		!lstrcmpi(ext, "gif") || !lstrcmpi(ext, "pal") || !lstrcmpi(ext, "pcx") || !lstrcmpi(ext, "png") || 
		!lstrcmpi(ext, "pbm") || !lstrcmpi(ext, "pgm") || !lstrcmpi(ext, "ppm") || !lstrcmpi(ext, "tga") ||
		!lstrcmpi(ext, "mng") || !lstrcmpi(ext, "tif") || !lstrcmpi(ext, "tiff") || !lstrcmpi(ext, "wmf") ||
		!lstrcmpi(ext, "jng") || !lstrcmpi(ext, "bmp") || !lstrcmpi(ext, "miff") || !lstrcmpi(ext, "pnm") ||
		!lstrcmpi(ext, "dds") || !lstrcmpi(ext, "psd") || !lstrcmpi(ext, "dpx") || !lstrcmpi(ext, "svg") ||
		!lstrcmpi(ext, "jpc") || !lstrcmpi(ext, "pdf") || !lstrcmpi(ext, "eps") || !lstrcmpi(ext, "ps") ||
		!lstrcmpi(ext, "exr") || !lstrcmpi(ext, "cr2") || !lstrcmpi(ext, "dcr") || !lstrcmpi(ext, "dng") ||
		!lstrcmpi(ext, "erf") || !lstrcmpi(ext, "mef") || !lstrcmpi(ext, "mrw") || !lstrcmpi(ext, "nef") ||
		!lstrcmpi(ext, "orf") || !lstrcmpi(ext, "pef") || !lstrcmpi(ext, "raf") || !lstrcmpi(ext, "raw") ||
		!lstrcmpi(ext, "sr2") || !lstrcmpi(ext, "x3f") || !lstrcmpi(ext, "3fr")))
	{
		env->ThrowError("ImmaRead: format (%s) not supported!", ext);
	}

	pPath = getenv("MAGICK_HOME");
	if (magick_home!=NULL) {
		char* cpath;
		int len = 12 + strlen(magick_home) + 1;
		cpath = (char *)malloc(len);
		strcpy(cpath, "MAGICK_HOME=");
		strcat(cpath, magick_home);
		int stat = putenv(cpath); // stat==0: success
		if (stat==-1) {
			env->ThrowError("ImmaRead: failed to set MAGICK_HOME variable, do this manually!");
		}
	} else {
		if (pPath==NULL) {
			env->ThrowError("ImmaRead: you haven't set the MAGICK_HOME variable and you didn't provide it as argument!");
		}
	}

	SplitFilename(fullname, img_path, img_name);
	_snprintf(img_name2, sizeof img_name2, fullname.c_str(), start); // first image
/*
	if (FAILED(__HrLoadAllImportsForDll("CORE_RL_MAGICK_.DLL"))) {
		env->ThrowError("failed on snap load, exiting\n");
	}
*/
	// Initialize:
	ExceptionInfo *exception;
	Image *image, *images;
	ImageInfo *image_info;
	MagickCoreGenesis(img_name2, MagickTrue);
	exception = AcquireExceptionInfo();
	
	// Get the properties of the image:
	image_info = CloneImageInfo((ImageInfo *) NULL);

	// printing and checking files
	if (strpbrk(img_name.c_str(), "*?")==NULL)  {
		fs.PrintFiles(img_path.c_str(), img_name.c_str(), start, end, gapless);
	} else {
		gapless = true;
		//searching for files
		fs.SearchFiles(img_path.c_str(), img_name.c_str());
	}

	if (gapless) {
		if (fs.GetFilecount()==0) {
			env->ThrowError("ImmaRead: no picture files found");
			return;
		}
	}

	_snprintf(img_name2, sizeof img_name2, fs.GetFilename(0)); // first existing image (doesn't need to exist for gapless = false)

	if (!lstrcmpi(ext, "svg")) {
		force_noresize = true;
	} else {
		force_noresize = false;
	}

	if ((force_size_x>0) && (force_size_y>0) && (force_noresize)) {
		itoa(force_size_x, fsize_x, 10);
		itoa(force_size_y, fsize_y, 10);
		strcpy(bufsize, fsize_x);
		strcat(bufsize, "x");
		strcat(bufsize, fsize_y);
		(void) CloneString(&image_info->size, bufsize);
	}

	(void) strcpy(image_info->filename, img_name2); // first image

	// setting density for vectorized images
	sprintf(dens, "%d", _density);
	(void) CloneString(&image_info->density, dens);

	images = PingImage(image_info, exception);
	// PingImage returns only the properties (thus not the pixels) of the image list,

	if (images == (Image *) NULL)
		env->ThrowError("ImmaRead: couldn't open picture file: severity=%x, reason=%s, description=%s", 
		exception->severity,
		exception->reason,
		exception->description);

	image = GetImageFromList(images, 0);

	// Set AviSynth values to the properties of the first image:
	if ((force_size_x>0) && (force_size_y>0) && !(force_noresize)) {
		vi.width = force_size_x;
		vi.height = force_size_y;
	} else {
		vi.width = image->columns;
		vi.height = image->rows;
	}

	if (strcmp(img_name2, fullname.c_str()) == 0) { // single image or animation
		if (animation) {
			vi.num_frames = GetImageListLength(images); // number of images in animated file
		} else {
			vi.num_frames = 1;
		}
		image_seq = false;
	} else {
		vi.num_frames = fs.GetFilecount(); //end - start + 1;
		image_seq = true;
	}

	// Colorspace check:
	// IsGrayImage(image,exception) == MagickTrue for Grayscale images.
	// But my test grayscale images return RGBColorspace instead of GRAYColorspace.
	vi.pixel_type = VideoInfo::CS_BGR32;

	// The other values are fixed for now:
	vi.SetFieldBased(false);
	vi.audio_samples_per_second = 0;

	// ticks = 100 / fps
	if ((animation)&&(vi.num_frames>1)) { // list (set fps from animation)
		int ticks = image->delay; // = nr of ticks (100 ticks = 1 sec) between images in the list
		vi.fps_denominator = 1000;
		if (ticks>0) {
			vi.fps_numerator = vi.fps_denominator*100/ticks;
		} else { // set fps=25 if ticks=0
			vi.fps_numerator = vi.fps_denominator*25;
		}
	} else { // set fps=25
		vi.fps_denominator = 1000;
		vi.fps_numerator = vi.fps_denominator*25;
	}

	// clean up memory:
	DestroyImageList(images);
	DestroyImageInfo(image_info);
	DestroyExceptionInfo(exception);
}

ImmaRead::~ImmaRead() {
	MagickCoreTerminus();
}

PVideoFrame __stdcall ImmaRead::GetFrame(int n, IScriptEnvironment* env) {
	PVideoFrame dst = env->NewVideoFrame(vi);
    unsigned char* dstp = dst->GetWritePtr();
    const int dst_pitch = dst->GetPitch();
    const int row_size = dst->GetRowSize();
	int width, height, width_orig, height_orig, width_new, height_new;
	char img_name2[MaxTextExtent];
	char BUF[256];
	double scale_fact = 1.0f;
	int offset_x = 0;
	int offset_y = 0;

	width = vi.width; // width of first image or forced width when specified
	height = vi.height; // height of first image or forced height when specified

	_snprintf(img_name2, sizeof img_name2, fs.GetFilename(n));
	SplitFilename(img_name2, img_path, img_name);

	// Initialize:
	ExceptionInfo *exception;
	Image *image, *images, *images2;
	ImageInfo *image_info;

	exception = AcquireExceptionInfo();
      
	// get the image:
	image_info = CloneImageInfo((ImageInfo *) NULL);

	if (force_noresize && (force_size_x>0) && (force_size_y>0)) {
		(void) CloneString(&image_info->size, bufsize);
	}

	(void) strcpy(image_info->filename, img_name2);

	// set density
	(void) CloneString(&image_info->density, dens);

	images = ReadImage(image_info, exception);

    // Check existence of image:
	if (images == (Image *) NULL) { // image doesn't exist
		memset(dstp, 0, dst_pitch * height);
		sprintf(BUF, "ImmaRead: couldn't open picture file:\n reason=%s\n", exception->reason);
		ApplyMessage(&dst, vi, BUF, width/4, 0xf0f080,0,0 , env);
		DestroyImageList(images);
		DestroyImageInfo(image_info);
		DestroyExceptionInfo(exception);
		MagickCoreTerminus();
		return dst;
	}

	if (animation) {
		images2 = CoalesceImages(images, exception);
		image = GetImageFromList(images2, n);
	} else {
		image = GetImageFromList(images, 0);
	}

	if (!(image->colorspace == RGBColorspace)) { //CMYK images
		SetImageType(image, TrueColorType);
		SetImageColorspace(image, RGBColorspace); 
	}

	width_orig = image->columns;
	height_orig = image->rows;

	if ((force_size_x>0) && (force_size_y>0) && !(force_noresize)) {
		memset(dstp, 0, dst_pitch * force_size_y);
		scale_fact = min(double(force_size_x)/double(width_orig),double(force_size_y)/double(height_orig));
	}

	width_new = scale_fact * width_orig;
	height_new = scale_fact * height_orig;

	// should load the image without modifying its geometry for non-svg images and resize afterwards:
	// http://www.imagemagick.org/api/resize.php
	// http://www.imagemagick.org/script/command-line-options.php?#filter
	if ((force_size_x>0) && (force_size_y>0) && !(force_noresize)) {
		offset_x = (force_size_x - width_new)>>1;
		offset_y = (force_size_y - height_new)>>1;
		FilterTypes ResizeMethod = MitchellFilter;
		image = ResizeImage(image, width_new, height_new, ResizeMethod, 1.0, exception);
	} else {
		// When loading image_seq: check width and height:
		if (image_seq) {
			if (height_new != height) {
				memset(dstp, 0, dst_pitch * height);
				sprintf(BUF, "ImmaRead: image %s\nmust have the same dimensions as the first image\n", img_name.c_str());
				ApplyMessage(&dst, vi, BUF, width/3, 0xf0f080,0,0 , env);
				// Cleanup
				DestroyImageList(images);
				if (animation) {
					DestroyImageList(images2);
				}
				DestroyImageInfo(image_info);
				DestroyExceptionInfo(exception);
				MagickCoreTerminus();
				return dst;
			}
			if (width_new != width) {
				memset(dstp, 0, dst_pitch * height);       
				sprintf(BUF, "ImmaRead: image %s\nmust have the same dimensions as the first image\n", img_name.c_str());
				ApplyMessage(&dst, vi, BUF, width/3, 0xf0f080,0,0 , env);
				// Cleanup
				DestroyImageList(images);
				if (animation) {
					DestroyImageList(images2);
				}
				DestroyImageInfo(image_info);
				DestroyExceptionInfo(exception);
				MagickCoreTerminus();
				return dst;
			}
		}
	}

	// transfer the pixels from the image to the frame:
	register const PixelPacket *p;  
	dstp += offset_y*(dst->GetPitch());
	for (int y=0; y < height_new; y++) {
		p = AcquireImagePixels(image, 0, height_new-1-y, width_new, 1, exception);
		// GetAuthenticPixels - http://www.imagemagick.org/api/cache.php new ???
		// p = GetImagePixels(image, 0, height-1-y, image->columns, 1);
		int z=0;
        for (int x=offset_x; x < (offset_x + width_new); x++) {
			z = 4*x;
			dstp[z]=p->blue;
			dstp[z+1]=p->green;
			dstp[z+2]=p->red;
			dstp[z+3]=255-p->opacity;
			p++;
        }
		dstp += dst->GetPitch();
	}

	// clean up memory:
	DestroyImageList(images);
	if (animation) {
		DestroyImageList(images2);
	}
	DestroyImageInfo(image_info);
	DestroyExceptionInfo(exception);
	MagickCoreTerminus();

	// overlay on video output: progress indicator
	if (info) {
		ostringstream text;
		if ((force_size_x>0) && (force_size_y>0) && !(force_noresize)) {
			text << "Frame " << n << ". Read from\n\"" << img_name2 << "\"\nOriginal size: " << width_orig << "x" << height_orig << "\nScaled size: " << width_new << "x" << height_new;
		} else {
			text << "Frame " << n << ". Read from\n\"" << img_name2 << "\"";
		}
		ApplyMessage(&dst, vi, text.str().c_str(), width/3, 0xf0f080,0,0 , env);
	}

	return dst;
}

AVSValue __cdecl Create_ImmaRead(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new ImmaRead(args[0].AsString(),
						args[1].AsInt(0),
						args[2].AsInt(100),
						args[3].AsInt(0),
						args[4].AsInt(0),
						args[5].AsBool(false),
						args[6].AsBool(false),
						args[7].AsInt(100),
						args[8].AsBool(false),
						args[9].AsString(),
						env);
}



class ImmaWrite : public GenericVideoFilter {
public:
	ImmaWrite(PClip _child, const char* _base_name, const int _start, const int _end, const char* _ext,
				const char* _mode, bool _animation, bool _use_dither, int _colors, bool _planar, int _density,
				const char* _compression, bool _info, const char* _magick_home, IScriptEnvironment* env);
	~ImmaWrite();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
private: // http://www.imagemagick.org/script/command-line-options.php?#type

	string base_name;     // base_name (supplied to the ctor)
	string old_path;      // working directory (location of script); *may change* during runtime
	string img_path;      // new directory (location of image) - path part of base_name
	string img_base_name; // image base filename (fname part of base_name)
	string extension;     // image extension (without leading .)

	const int start, end;
	bool animation, use_dither, planar, info;
	int mode, colors, compression;
	const char* magick_home;
	char* chroma_sampling;
	char dens[10];
	string name;
	bool raw, y4m;
	FILE *f;
	char fname[512];
	int status;

	enum { TrueColor = 0, Bilevel = 1, Grayscale = 2, Palette = 3,
		TrueColorMatte = 4, BilevelMatte = 5, GrayscaleMatte = 6, PaletteMatte = 7 };
	enum { NoCompression = 0, BZipCompression = 1, FaxCompression = 2, Group4Compression = 3,
		JPEGCompression = 4, LZWCompression = 5, RLECompression = 6, ZipCompression = 7 };

	Image *images;
	int nr_frames;
	float ticks;

	// returns the filename (without path part) of the image to store for frame n
	string FrameName(int n) const;
};

ImmaWrite::ImmaWrite(PClip _child, const char* _base_name, const int _start, const int _end, const char* _ext,
					 const char* _mode, bool _animation, bool _use_dither, int _colors, bool _planar, int _density,
					 const char* _compression, bool _info, const char* _magick_home, IScriptEnvironment* env) :
	GenericVideoFilter(_child), base_name(_base_name), extension(_ext), start(_start), end(_end),
		animation(_animation), use_dither(_use_dither), colors(_colors), planar(_planar), info(_info),
		magick_home(_magick_home) {
	
	const char* ext = extension.c_str();
	if (!(!lstrcmpi(ext, "jpeg") || !lstrcmpi(ext, "jpe") || !lstrcmpi(ext, "jpg") || !lstrcmpi(ext, "jp2") ||
		!lstrcmpi(ext, "gif") || !lstrcmpi(ext, "pal") || !lstrcmpi(ext, "pcx") || !lstrcmpi(ext, "png") ||
		!lstrcmpi(ext, "pbm") || !lstrcmpi(ext, "pgm") || !lstrcmpi(ext, "ppm") || !lstrcmpi(ext, "tga") ||
		!lstrcmpi(ext, "mng") || !lstrcmpi(ext, "tif") || !lstrcmpi(ext, "tiff") || !lstrcmpi(ext, "psd") ||
		!lstrcmpi(ext, "jng") || !lstrcmpi(ext, "bmp") || !lstrcmpi(ext, "miff") || !lstrcmpi(ext, "pnm") ||
		!lstrcmpi(ext, "dds") || !lstrcmpi(ext, "dpx") || !lstrcmpi(ext, "raw") || !lstrcmpi(ext, "y4m") ||
		!lstrcmpi(ext, "jpc") || !lstrcmpi(ext, "bmp3") || !lstrcmpi(ext, "bmp4") || !lstrcmpi(ext, "pdf") ||
		!lstrcmpi(ext, "eps") || !lstrcmpi(ext, "ps"))) {
		env->ThrowError("ImmaWrite: format (%s) not supported!", ext);
	}

	if ( !(!lstrcmpi(ext, "gif") || !lstrcmpi(ext, "mng") || !lstrcmpi(ext, "ppm") || !lstrcmpi(ext, "psd") ||
		!lstrcmpi(ext, "tif") || !lstrcmpi(ext, "tiff") || !lstrcmpi(ext, "raw") || !lstrcmpi(ext, "y4m") ||
		!lstrcmpi(ext, "pdf") || !lstrcmpi(ext, "ps")) && animation ) {
		env->ThrowError("ImmaWrite: format (%s) not supported as animation!", ext);
	}

	raw = false;
	if (!lstrcmpi(ext, "raw") || !lstrcmpi(ext, "y4m")) {
		raw = true;
	}

	if (vi.IsYUV() && (!(raw))) {
		env->ThrowError("ImmaWrite: your source is YCbCr, but only \"raw\" and \"y4m\" are supported for this colorspace!");
	}

	y4m = false;
	if (!lstrcmpi(ext, "y4m")) {
		y4m = true;
	}

	// chroma_sampling =
	//  420mpeg2
	//  411 // v2.60
	//  422
	//  444 // v2.60
	//	444alpha // not supported
	//  mono // greyscale=true
	if (y4m) {
		if (vi.IsYUV() && (lstrcmpi(_mode, "Grayscale") == 0)) { // greyscale
			chroma_sampling = new char[5];
			chroma_sampling = "mono";
			planar = true;
		} else if (vi.IsYV12()) {
			chroma_sampling = new char[9];
			chroma_sampling = "420mpeg2";
		} else if (vi.IsYUY2()) {
			chroma_sampling = new char[4];
			chroma_sampling = "422";
			planar = true;
		} else { // RGB(A)
			env->ThrowError("ImmaWrite: your source is RGB, but only YCbCr is supported for y4m!");
		}
	}

	// setting density
	sprintf(dens, "%d", _density);

	// check bounds
	if (end>vi.num_frames-1)
		env->ThrowError("ImmaWrite: make sure that end<framecount(clip)");
	if (start>end)
		env->ThrowError("ImmaWrite: make sure that start<=end");

	if (vi.IsRGB24()) {
		if (lstrcmpi(_mode, "TrueColor") == 0) // default
			mode = TrueColor;
		else if (lstrcmpi(_mode, "Bilevel") == 0)
			mode = Bilevel;
		else if (lstrcmpi(_mode, "Grayscale") == 0) {
			mode = Grayscale;
			planar = true;
		}
		else if (lstrcmpi(_mode, "Palette") == 0)
			mode = Palette;
		else 
			env->ThrowError("ImmaWrite: mode must be either \"TrueColor\", \"Bilevel\", \"Grayscale\" or \"Palette\"!");
	} else if (vi.IsRGB32()) {
		if (lstrcmpi(_mode, "TrueColor") == 0) // default
			mode = TrueColorMatte;
		else if (lstrcmpi(_mode, "Bilevel") == 0)
			mode = Bilevel; // Bilevel with alpha channel not supported
		else if (lstrcmpi(_mode, "Grayscale") == 0) {
			mode = GrayscaleMatte;
			planar = true;
		}
		else if (lstrcmpi(_mode, "Palette") == 0)
			mode = PaletteMatte;
		else 
			env->ThrowError("ImmaWrite: mode must be either \"TrueColor\", \"Bilevel\", \"Grayscale\" or \"Palette\"!");	
	} else { // YCbCr [used when saving raw files]
		if (lstrcmpi(_mode, "TrueColor") == 0) // default
			mode = TrueColor;
		else if (lstrcmpi(_mode, "Grayscale") == 0) {
			mode = Grayscale;
			planar = true;
		}
		else 
			env->ThrowError("ImmaWrite: mode must be either \"TrueColor\", \"Bilevel\", \"Grayscale\" or \"Palette\"!");	
	}

	// converts clip to greyscale => will be saved as a greyscale image by ImageMagick:
	if ((mode == Grayscale)||(mode == GrayscaleMatte)) {
		try 
		{
			AVSValue cache_args[1] = { child };
			child = env->Invoke("GreyScale", AVSValue(cache_args,1)).AsClip();
		}
		catch (AvisynthError)
		{
			env->ThrowError("ImmaWrite: error invoking GreyScale!");
		}
	}

	if (lstrcmpi(_compression, "NoCompression") == 0) // default
		compression = NoCompression;
	else if (lstrcmpi(_compression, "BZipCompression") == 0)
		compression = BZipCompression;
	else if (lstrcmpi(_compression, "FaxCompression") == 0)
		compression = FaxCompression;
	else if (lstrcmpi(_compression, "Group4Compression") == 0)
		compression = Group4Compression;
	else if (lstrcmpi(_compression, "JPEGCompression") == 0)
		 compression = JPEGCompression;
	else if (lstrcmpi(_compression, "LZWCompression") == 0)
		compression = LZWCompression;
	else if (lstrcmpi(_compression, "RLECompression") == 0)
		compression = RLECompression;
	else if (lstrcmpi(_compression, "ZipCompression") == 0)
		compression = ZipCompression;
	else 
		env->ThrowError("ImmaWrite: compression must be either \"xxx\" or \"xxx\"!");	

	char* pPath;
	pPath = getenv("MAGICK_HOME");
	if (magick_home!=NULL) {
		char* cpath;
		int len = 12 + strlen(magick_home) + 1;
		cpath = (char *)malloc(len);
		strcpy(cpath, "MAGICK_HOME=");
		strcat(cpath, magick_home);
		int stat = putenv(cpath); // stat==0: success
		if (stat==-1) {
			env->ThrowError("ImmaWrite: failed to set MAGICK_HOME variable, do this manually!");
		}
		free(cpath);
	} else {
		if (pPath==NULL) {
			env->ThrowError("ImmaWrite: you haven't set the MAGICK_HOME variable and you didn't provide the argument!");
		}
	}

	if (animation) // create full name of image file (it is one file)
		name = base_name + "." + extension; 
	else           // construct a full name of image file assuming frame 0
		name = FrameName(0);

	// initialize:
	// for (animation==true) there is no need to re-supply name
	MagickCoreGenesis(name.c_str(), MagickTrue);

	// use a list when writing animations:
	if (animation) {
		images = NewImageList();
		nr_frames = vi.num_frames;
		// the number of ticks (= image->delay) and the fact that there are 100 ticks in one second
		// determines the time between two images.
		// delay = x ticks and this implies x/100 seconds between two images.
		// this relates to the framerate: 1/fps = x/100 and thus x ticks = 100/fps
		ticks = 100*float(vi.fps_denominator)/float(vi.fps_numerator);
		if (raw) {
			strcpy(fname, name.c_str());
			f = fopen(fname, "wb");
			if (f == NULL) {
				env->ThrowError("ImmaWrite: couldn't open file!");
			} else {
				status = 1;
			}
			if (y4m) {
				fprintf(f, "YUV4MPEG2 W%d H%d C%s F%ld:%ld Ip A0:0\n",
					vi.width, vi.height, chroma_sampling, vi.fps_numerator, vi.fps_denominator);
			}
		}
	}
}

ImmaWrite::~ImmaWrite() {
	MagickCoreTerminus();
	delete[] chroma_sampling;
}

string ImmaWrite::FrameName(int n) const {
	ostringstream oss;
	oss << base_name << setfill('0') << setw(6) << n << '.' << extension;
	return string(oss.str());
}

PVideoFrame __stdcall ImmaWrite::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
		
	const BYTE* srcp = src->GetReadPtr();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();

	ExceptionInfo *exception;
	Image *image;
	ImageInfo *image_info;
	QuantizeInfo *quantize_info; // stuff with quantize_info enables us to write PNG greyscale pics
	PixelPacket *q;
	int x, y;
		
	image_info = CloneImageInfo((ImageInfo *) NULL);
	quantize_info = CloneQuantizeInfo((QuantizeInfo *) NULL);
	exception = AcquireExceptionInfo();
	image = AllocateImage(image_info);
	SetImageExtent(image, vi.width, vi.height); // creates canvas with specified size

	int planes, z;
	if (vi.IsPlanar()) {
		planar=true;
	}
	if (planar) {
		if (vi.IsRGB32()) {
			planes=4;
		} else { // RGB24, YV16, YV12
			planes=3;
		}
	}

	if (start <= n && n <= end) {
		if (raw) {
			if (!(animation)) {
				strcpy(fname, FrameName(n).c_str());
				f = fopen(fname, "wb");
				if (f == NULL) {// couldn't open file
					env->MakeWritable(&src);
					ostringstream text;
					text << "ImmaWrite: couldn't open file!";
					ApplyMessage(&src, vi, text.str().c_str(), vi.width/3, 0xf0f080,0,0 , env);
					return src; // write error on frame
				} else {
					status = 1;
				}
				if (y4m) {
					fprintf(f, "YUV4MPEG2 W%d H%d C%s F%ld:%ld Ip A0:0\n",
						vi.width, vi.height, chroma_sampling, vi.fps_numerator, vi.fps_denominator);
				}
			}

			if (y4m) {
				fwrite("FRAME\n", 1, 6, f);
			}

			switch (mode)
			{
			case TrueColor:	case TrueColorMatte:
				{
					if (planar) {
						if (vi.IsRGB()) { // save to planar BGR(A)
							for (z=0; z<planes; z++) {
								srcp = src->GetReadPtr();
								for (y=0; y<src_height; y++) {
									for (x=z; x<src_width; x+=planes) {
										fwrite(&srcp[x], 1, 1, f);
									}
									srcp += src_pitch;
								}
							}
						} else if (vi.IsYUY2()) { // save to planar YUV - (YV16) (remove in v2.6)
							// luma (Y)
							srcp = src->GetReadPtr();
							for (y=0; y<src_height; y++) {
								for (x=0; x<src_width; x+=4) {
									fwrite(&srcp[x], 1, 1, f);
									fwrite(&srcp[x+2], 1, 1, f);
								}
								srcp += src_pitch;
							}
							// chroma (U)
							srcp = src->GetReadPtr();
							for (y=0; y<src_height; y++) {
								for (x=0; x<src_width; x+=4) {
									fwrite(&srcp[x+1], 1, 1, f);
								}
								srcp += src_pitch;
							}
							// chroma (V)
							srcp = src->GetReadPtr();
							for (y=0; y<src_height; y++) {
								for (x=0; x<src_width; x+=4) {
									fwrite(&srcp[x+3], 1, 1, f);
								}
								srcp += src_pitch;
							}
						} else { // YV12
							static const int plane[] = {PLANAR_Y, PLANAR_U, PLANAR_V};
	
							for (int p=0; p<3; p++) {
								int w = src_width >> (p ? 1 : 0); // adjust in v2.6
								int h = src_height >> (p ? 1 : 0); // adjust in v2.6
								const BYTE* srcp = src->GetReadPtr(plane[p]);
								int src_pitch = src->GetPitch(plane[p]);
								for (y=0; y<h; y++) {
									fwrite(srcp, 1, w, f);
									srcp += src_pitch;
								}
							}
						}
					} else { // interleaved (RGB, YUY2)
						for (y=0; y<src_height; y++) {
							fwrite(srcp, 1, src_width, f);
							srcp += src_pitch;
						}
					}
					break;
				}
			case Grayscale:
				{
					if (vi.IsRGB()) {
						srcp += src_pitch*(src_height-1); // write upside down
						for (y=0; y<src_height; y++) {
							for (x=0; x<src_width; x+=planes) {
								fwrite(&srcp[x], 1, 1, f);
							}
							srcp -= src_pitch;
						}
					} else if (vi.IsYUY2()) {						
						// save luma (Y)
						srcp = src->GetReadPtr();
						for (y=0; y<src_height; y++) {
							for (x=0; x<src_width; x+=4) {
								fwrite(&srcp[x], 1, 1, f);
								fwrite(&srcp[x+2], 1, 1, f);
							}
							srcp += src_pitch;
						}
					} else { // YV12
						const BYTE* srcp = src->GetReadPtr(PLANAR_Y);
						int src_pitch = src->GetPitch(PLANAR_Y);
						int w = src_width;
						int h = src_height;
						for (y=0; y<h; y++) {
							fwrite(srcp, 1, w, f);
							srcp += src_pitch;
						}
					}
					break;
				}
			}
			
			if (animation) {
				if (n==end) {
					fclose(f);
				}
			} else {
				fclose(f);
			}
		} else { // writing non-raw images
			if (vi.IsRGB32()) {
				SetImageAlphaChannel(image, ActivateAlphaChannel); // activates the alpha channel
				for (y=src_height-1; y>=0; --y)	{ // write pixels upside down
					// http://www.imagemagick.org/script/architecture.php
					q = GetImagePixels(image, 0, y, vi.width, 1);
					if (q == (PixelPacket *) NULL)
						break;
					for (x=0; x < vi.width; x++) {
						q->blue=srcp[4*x];
						q->green=srcp[4*x+1];
						q->red=srcp[4*x+2];
						q->opacity=255-srcp[4*x+3];
						q++;
					}
					srcp += src_pitch;
					if (SyncImagePixels(image) == MagickFalse)
						break;
				}
			} else { // RGB24
				for (y=src_height-1; y>=0; --y)	{ // write pixels upside down
					// http://www.imagemagick.org/script/architecture.php
					q = GetImagePixels(image, 0, y, vi.width, 1);
					if (q == (PixelPacket *) NULL)
						break;
					for (x=0; x < vi.width; x++) {
						q->blue=srcp[3*x];
						q->green=srcp[3*x+1];
						q->red=srcp[3*x+2];
						q++;
					}
					srcp += src_pitch;
					if (SyncImagePixels(image) == MagickFalse)
						break;
				}
			}
			
			if (animation) {
				strcpy(image->filename, name.c_str());
			} else {
				strcpy(image->filename, FrameName(n).c_str());
			}

			switch (mode)
			{
			case Bilevel:
				{
					SetImageType(image, BilevelType);
					break;
				}
			case Grayscale:
				{
					//	SetImageColorspace(image, GRAYColorspace); // converts to GRAY
					SetImageType(image, GrayscaleType);
					quantize_info->colorspace = GRAYColorspace;
					break;
				}
			case Palette:
				{
					SetImageType(image, PaletteType);
					break;
				}
			case TrueColor:
				{
					SetImageType(image, TrueColorType);
					break;
				}
			case GrayscaleMatte:
				{
					//	SetImageColorspace(image, GRAYColorspace); // converts to GRAY
					SetImageType(image, GrayscaleMatteType);
					quantize_info->colorspace = GRAYColorspace;
					break;
				}
			case PaletteMatte:
				{
					SetImageType(image, PaletteMatteType);
					break;
				}
			case TrueColorMatte:
				{
					SetImageType(image, TrueColorMatteType);
					break;
				}
			}

			switch (compression)
			{
			case NoCompression:
				{
					image_info->compression = (CompressionType)NoCompression;
					break;
				}
			case BZipCompression:
				{
					image_info->compression = (CompressionType)BZipCompression;
					break;
				}
			case FaxCompression:
				{
					image_info->compression = (CompressionType)FaxCompression;
					break;
				}
			case Group4Compression:
				{
					image_info->compression = (CompressionType)Group4Compression;
					break;
				}
			case JPEGCompression:
				{
					image_info->compression = (CompressionType)JPEGCompression;
					break;
				}
			case LZWCompression:
				{
					image_info->compression = (CompressionType)LZWCompression;
					break;
				}
			case RLECompression:
				{
					image_info->compression = (CompressionType)RunlengthEncodedCompression;
					break;
				}
			case ZipCompression:
				{
					image_info->compression = (CompressionType)ZipCompression;
					break;
				}
			}

			if (colors>0) { // color reduction
				quantize_info->tree_depth = 8;
				quantize_info->number_colors = colors;
			}

			if (!(use_dither)) { // turn off dithering
				quantize_info->dither = MagickFalse;
			}

			if ( (colors>0) || !(use_dither) || (mode==Grayscale) || (mode==GrayscaleMatte) ) {
				// QuantizeImage analyzes the colors within a reference image and chooses a fixed number of colors
				// to represent the image
				QuantizeImage(quantize_info, image);
			}

			// set density (default=72)
			(void) CloneString(&image_info->density, dens);

			if (animation) {
				image->delay = int(ticks);
				AppendImageToList(&images, image);
				// writes animation to hard-drive when the last frame is added to the list
				if (n==end) { 
					status = WriteImage(image_info, images);
					if (status == 0) {// couldn't write animation
						env->MakeWritable(&src);
						ostringstream text;
						text << "Couldn't write animation to \"" << name.c_str() << "\"";
						ApplyMessage(&src, vi, text.str().c_str(), vi.width/3, 0xf0f080,0,0 , env);
					}
					// clean up memory
					DestroyImageList(images);
				}
			} else { // write image to hard-drive
				status = WriteImage(image_info, image);
				if (status == 0) {// couldn't write image
					env->MakeWritable(&src);
					ostringstream text;
					text << "Couldn't write image to \"" << FrameName(n).c_str() << "\"";
					ApplyMessage(&src, vi, text.str().c_str(), vi.width/3, 0xf0f080,0,0 , env);
				}
				DestroyImage(image);
			}

			// clean up memory
			DestroyImageInfo(image_info);
			DestroyExceptionInfo(exception);
		}
	} else {
		return src;
	}

	if ( (info) && (status>0) ) { // overlay on video output: progress indicator
		if (start <= n && n <= end) {
			if (animation) {
				if (n==end) { 
					env->MakeWritable(&src);
					ostringstream text;
					text << "Animation (frames: " << start << "-" << end << ") written to \"" << name.c_str() << "\"";
					ApplyMessage(&src, vi, text.str().c_str(), vi.width/4, 0xf0f080,0,0 , env);
				}
			} else {
				env->MakeWritable(&src);
				ostringstream text;
				text << "Frame " << n << " written to \"" << FrameName(n).c_str() << "\"";
				ApplyMessage(&src, vi, text.str().c_str(), vi.width/4, 0xf0f080,0,0 , env);
			}
		}
	}

	return src;
}

AVSValue __cdecl Create_ImmaWrite(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new ImmaWrite(args[0].AsClip(),
						 env->SaveString(args[1].AsString("c:\\")),
                         args[2].AsInt(0),
						 args[3].AsInt((args[0].AsClip()->GetVideoInfo().num_frames)-1),
						 env->SaveString(args[4].AsString("tif")),
						 args[5].AsString("TrueColor"),
						 args[6].AsBool(false),
						 args[7].AsBool(true),
						 args[8].AsInt(0),
						 args[9].AsBool(false),
						 args[10].AsInt(100),
						 args[11].AsString("NoCompression"),
 						 args[12].AsBool(false),
						 args[13].AsString(),
						 env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("ImmaRead", "[filename]s[start]i[end]i[size_x]i[size_y]i[animation]b[gapless]b[density]i[info]b[magick_home]s", Create_ImmaRead, 0);
    env->AddFunction("ImmaWrite", "c[filename]s[start]i[end]i[type]s[mode]s[animation]b[use_dither]b[colors]i[planar]b[density]i[compression]s[info]b[magick_home]s", Create_ImmaWrite, 0);
	return "ImageMagick plugin";
}
