// Avisynth v2.5.  Copyright 2002 Ben Rudiak-Gould et al.
// http://www.avisynth.org

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .
//
// Linking Avisynth statically or dynamically with other modules is making a
// combined work based on Avisynth.  Thus, the terms and conditions of the GNU
// General Public License cover the whole combination.
//
// As a special exception, the copyright holders of Avisynth give you
// permission to link Avisynth with independent modules that communicate with
// Avisynth solely through the interfaces defined in avisynth.h, regardless of the license
// terms of these independent modules, and to copy and distribute the
// resulting combined work under terms of your choice, provided that
// every copy of the combined work is accompanied by a complete copy of
// the source code of Avisynth (the version of Avisynth used to produce the
// combined work), being distributed under the terms of the GNU General
// Public License plus this exception.  An independent module is a module
// which is not derived from or based on Avisynth, such as 3rd-party filters,
// import and export plugins, or graphical user interfaces.


#include "stdafx.h"

#include "text-overlay.h"

#include<string>
#include<sstream>

using namespace std;



/******************************
 *******   Anti-alias    ******
 *****************************/

Antialiaser::Antialiaser(int width, int height, const char fontname[], int size,
   int _textcolor, int _halocolor, int font_width, int font_angle, bool _interlaced)
 : w(width), h(height), textcolor(_textcolor), halocolor(_halocolor), alpha_calcs(0),
   dirty(true), interlaced(_interlaced)
{
  struct {
    BITMAPINFOHEADER bih;
    RGBQUAD clr[2];
  } b;

  b.bih.biSize                    = sizeof(BITMAPINFOHEADER);
  b.bih.biWidth                   = width * 8 + 32;
  b.bih.biHeight                  = height * 8 + 32;
  b.bih.biBitCount                = 1;
  b.bih.biPlanes                  = 1;
  b.bih.biCompression             = BI_RGB;
  b.bih.biXPelsPerMeter   = 0;
  b.bih.biYPelsPerMeter   = 0;
  b.bih.biClrUsed                 = 2;
  b.bih.biClrImportant    = 2;
  b.clr[0].rgbBlue = b.clr[0].rgbGreen = b.clr[0].rgbRed = 0;
  b.clr[1].rgbBlue = b.clr[1].rgbGreen = b.clr[1].rgbRed = 255;

  if (hdcAntialias = CreateCompatibleDC(NULL)) {
	if (hbmAntialias = CreateDIBSection
	  ( hdcAntialias,
		(BITMAPINFO *)&b,
		DIB_RGB_COLORS,
		&lpAntialiasBits,
		NULL,
		0 )) {
	  hbmDefault = (HBITMAP)SelectObject(hdcAntialias, hbmAntialias);

	  HFONT newfont = LoadFont(fontname, size, true, false, font_width, font_angle);
	  hfontDefault = newfont ? (HFONT)SelectObject(hdcAntialias, newfont) : 0;

	  SetMapMode(hdcAntialias, MM_TEXT);
	  SetTextColor(hdcAntialias, 0xffffff);
	  SetBkColor(hdcAntialias, 0);

	  alpha_calcs = new unsigned short[width*height*4];

	  if (!alpha_calcs) FreeDC();
	}
  }
}


Antialiaser::~Antialiaser() {
  FreeDC();
  if (alpha_calcs) delete[] alpha_calcs;
}


HDC Antialiaser::GetDC() {
  dirty = true;
  return hdcAntialias;
}


void Antialiaser::FreeDC() {
  if (hdcAntialias) {
    if (hbmDefault) {
	  DeleteObject(SelectObject(hdcAntialias, hbmDefault));
	  hbmDefault = 0;
	}
    if (hfontDefault) {
	  DeleteObject(SelectObject(hdcAntialias, hfontDefault));
	  hfontDefault = 0;
	}
    DeleteDC(hdcAntialias);
    hdcAntialias = 0;
  }
}


void Antialiaser::Apply( const VideoInfo& vi, PVideoFrame* frame, int pitch)
{
  if (!alpha_calcs) return;

  if (vi.IsRGB32())
    ApplyRGB32((*frame)->GetWritePtr(), pitch);
  else if (vi.IsRGB24())
    ApplyRGB24((*frame)->GetWritePtr(), pitch);
  else if (vi.IsYUY2())
    ApplyYUY2((*frame)->GetWritePtr(), pitch);
  else if (vi.IsYV12())
    ApplyYV12((*frame)->GetWritePtr(), pitch,
              (*frame)->GetPitch(PLANAR_U),
			  (*frame)->GetWritePtr(PLANAR_U),
			  (*frame)->GetWritePtr(PLANAR_V) );
}


void Antialiaser::ApplyYV12(BYTE* buf, int pitch, int pitchUV, BYTE* bufU, BYTE* bufV) {
  if (dirty) {
    GetAlphaRect();
	xl &= -2; xr |= 1;
	yb &= -2; yt |= 1;
  }
  const int w4 = w*4;
  unsigned short* alpha = alpha_calcs + yb*w4;
  buf  += pitch*yb;
  bufU += (pitchUV*yb)>>1;
  bufV += (pitchUV*yb)>>1;

  for (int y=yb; y<=yt; y+=2) {
    for (int x=xl; x<=xr; x+=2) {
      const int x4 = x<<2;
      const int basealpha00 = alpha[x4+0];
      const int basealpha10 = alpha[x4+4];
      const int basealpha01 = alpha[x4+0+w4];
      const int basealpha11 = alpha[x4+4+w4];
      const int basealphaUV = basealpha00 + basealpha10 + basealpha01 + basealpha11;

      if (basealphaUV != 1024) {
        buf[x+0]       = (buf[x+0]       * basealpha00 + alpha[x4+3]   ) >> 8;
        buf[x+1]       = (buf[x+1]       * basealpha10 + alpha[x4+7]   ) >> 8;
        buf[x+0+pitch] = (buf[x+0+pitch] * basealpha01 + alpha[x4+3+w4]) >> 8;
        buf[x+1+pitch] = (buf[x+1+pitch] * basealpha11 + alpha[x4+7+w4]) >> 8;

        const int au  = alpha[x4+2] + alpha[x4+6] + alpha[x4+2+w4] + alpha[x4+6+w4];
        bufU[x>>1] = (bufU[x>>1] * basealphaUV + au) >> 10;

        const int av  = alpha[x4+1] + alpha[x4+5] + alpha[x4+1+w4] + alpha[x4+5+w4];
        bufV[x>>1] = (bufV[x>>1] * basealphaUV + av) >> 10;
      }
    }
    buf += pitch<<1;
    bufU += pitchUV;
    bufV += pitchUV;
    alpha += w<<3;
  }
}




void Antialiaser::ApplyYUY2(BYTE* buf, int pitch) {
  if (dirty) {
    GetAlphaRect();
	xl &= -2; xr |= 1;
  }
  unsigned short* alpha = alpha_calcs + yb*w*4;
  buf += pitch*yb;

  for (int y=yb; y<=yt; ++y) {
    for (int x=xl; x<=xr; x+=2) {
      const int basealpha0  = alpha[x*4+0];
      const int basealpha1  = alpha[x*4+4];
      const int basealphaUV = basealpha0 + basealpha1;

      if (basealphaUV != 512) {
        buf[x*2+0] = (buf[x*2+0] * basealpha0 + alpha[x*4+3]) >> 8;
        buf[x*2+2] = (buf[x*2+2] * basealpha1 + alpha[x*4+7]) >> 8;

        const int au  = alpha[x*4+2] + alpha[x*4+6];
        buf[x*2+1] = (buf[x*2+1] * basealphaUV + au) >> 9;

        const int av  = alpha[x*4+1] + alpha[x*4+5];
        buf[x*2+3] = (buf[x*2+3] * basealphaUV + av) >> 9;
      }
    }
    buf += pitch;
    alpha += w*4;
  }
}


void Antialiaser::ApplyRGB24(BYTE* buf, int pitch) {
  if (dirty) GetAlphaRect();
  unsigned short* alpha = alpha_calcs + yb*w*4;
  buf  += pitch*(h-yb-1);

  for (int y=yb; y<=yt; ++y) {
    for (int x=xl; x<=xr; ++x) {
      const int basealpha = alpha[x*4+0];
      if (basealpha != 256) {
        buf[x*3+0] = (buf[x*3+0] * basealpha + alpha[x*4+1]) >> 8;
        buf[x*3+1] = (buf[x*3+1] * basealpha + alpha[x*4+2]) >> 8;
        buf[x*3+2] = (buf[x*3+2] * basealpha + alpha[x*4+3]) >> 8;
      }
    }
    buf -= pitch;
    alpha += w*4;
  }
}


void Antialiaser::ApplyRGB32(BYTE* buf, int pitch) {
  if (dirty) GetAlphaRect();
  unsigned short* alpha = alpha_calcs + yb*w*4;
  buf  += pitch*(h-yb-1);

  for (int y=yb; y<=yt; ++y) {
    for (int x=xl; x<=xr; ++x) {
      const int basealpha = alpha[x*4+0];
      if (basealpha != 256) {
        buf[x*4+0] = (buf[x*4+0] * basealpha + alpha[x*4+1]) >> 8;
        buf[x*4+1] = (buf[x*4+1] * basealpha + alpha[x*4+2]) >> 8;
        buf[x*4+2] = (buf[x*4+2] * basealpha + alpha[x*4+3]) >> 8;
      }
    }
    buf -= pitch;
    alpha += w*4;
  }
}


void Antialiaser::GetAlphaRect() {

  dirty = false;

  static BYTE bitcnt[256],    // bit count
              bitexl[256],    // expand to left bit
              bitexr[256];    // expand to right bit
  static bool fInited = false;
  static unsigned short gamma[129]; // Gamma lookups

  if (!fInited) {
    fInited = true;
    int i;

    const double scale = 516*64/sqrt(128);
    for(i=0; i<=128; i++)
      gamma[i]=unsigned short(sqrt(i) * scale + 0.5); // Gamma = 2.0

    for(i=0; i<256; i++) {
      BYTE b=0, l=0, r=0;

      if (i&  1) { b=1; l|=0x01; r|=0xFF; }
      if (i&  2) { ++b; l|=0x03; r|=0xFE; }
      if (i&  4) { ++b; l|=0x07; r|=0xFC; }
      if (i&  8) { ++b; l|=0x0F; r|=0xF8; }
      if (i& 16) { ++b; l|=0x1F; r|=0xF0; }
      if (i& 32) { ++b; l|=0x3F; r|=0xE0; }
      if (i& 64) { ++b; l|=0x7F; r|=0xC0; }
      if (i&128) { ++b; l|=0xFF; r|=0x80; }

      bitcnt[i] = b;
      bitexl[i] = l;
      bitexr[i] = r;
    }
  }

  const int RYtext = ((textcolor>>16)&255), GUtext = ((textcolor>>8)&255), BVtext = (textcolor&255);
  const int RYhalo = ((halocolor>>16)&255), GUhalo = ((halocolor>>8)&255), BVhalo = (halocolor&255);

  // Scaled Alpha
  const int Atext = 255 - ((textcolor >> 24) & 0xFF);
  const int Ahalo = 255 - ((halocolor >> 24) & 0xFF);

  const int srcpitch = (w+4+3) & -4;

  xl=0;
  xr=w+1;
  yt=-1;
  yb=h;

  unsigned short* dest = alpha_calcs;
  for (int y=0; y<h; ++y) {
    BYTE* src = (BYTE*)lpAntialiasBits + ((h-y-1)*8 + 20) * srcpitch + 2;
    int wt = w;
    do {
      int i;

      DWORD tmp = interlaced;
      __asm {           // test if the whole area isn't just plain black
        mov edx, srcpitch
        mov esi, src
        mov ecx, edx
        dec esi
        shl ecx, 3
        sub esi, ecx  ; src - 8*pitch - 1
        cmp tmp,-1
        jnz do32

        lea edi,[esi+edx*2]
        xor eax,eax
        xor ecx,ecx
        jmp do24
do32:
        sar ecx, 1
        sub esi, ecx  ; src - 12*pitch - 1
        lea edi,[esi+edx*2]

        mov eax, [esi]  ; repeat 32 times
        mov ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
        or eax, [esi]
        or ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
do24:
        or eax, [esi]  ; repeat 24 times
        or ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
        or eax, [esi]
        or ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
        or eax, [esi]
        or ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
        or eax, [esi]
        or ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
        or eax, [esi]
        or ecx, [esi+edx]
        lea esi,[esi+edx*4]
        or eax, [edi]
        or ecx, [edi+edx]
        lea edi,[edi+edx*4]
        or eax, [esi]
        or ecx, [esi+edx]
        or eax, [edi]
        or ecx, [edi+edx]

        or eax, ecx
        and eax, 0x00ffffff
        mov tmp, eax
      }

      if (tmp != 0) {     // quick exit in a common case
		if (wt >= xl) xl=wt;
		if (wt <= xr) xr=wt;
		if (y  >= yt) yt=y;
		if (y  <= yb) yb=y;

        int alpha1, alpha2;

        alpha1 = alpha2 = 0;

		if (interlaced) {
		  BYTE topmask=0, cenmask=0, botmask=0;
		  BYTE hmasks[16], mask;
		  
		  for(i=-4; i<12; i++) {// For interlaced include extra half cells above and below
			mask = src[srcpitch*i];
			// How many lit pixels in the centre cell?
			alpha1 += bitcnt[mask];
			// turn on all halo bits if cell has any lit pixels
			mask = - !! mask;
			// Check left and right neighbours, extend the halo
			// mask 8 pixels in from the nearest lit pixels.
			mask |= bitexr[src[srcpitch*i-1]];
			mask |= bitexl[src[srcpitch*i+1]];
			hmasks[i+4] = mask;
		  }

		  // Extend halo vertically to 8x8 blocks
		  for(i=-4; i<4;  i++) topmask |= hmasks[i+4];
		  for(i=0;  i<8;  i++) cenmask |= hmasks[i+4];
		  for(i=4;  i<12; i++) botmask |= hmasks[i+4];
		  // Check the 3x1.5 cells above
		  for(mask = topmask, i=-4; i<4; i++) {
			mask |= bitexr[ src[srcpitch*(i+8)-1] ];
			mask |=    - !! src[srcpitch*(i+8)  ];
			mask |= bitexl[ src[srcpitch*(i+8)+1] ];
			hmasks[i+4] |= mask;
		  }
		  for(mask = cenmask, i=0; i<8; i++) {
			mask |= bitexr[ src[srcpitch*(i+8)-1] ];
			mask |=    - !! src[srcpitch*(i+8)  ];
			mask |= bitexl[ src[srcpitch*(i+8)+1] ];
			hmasks[i+4] |= mask;
		  }
		  for(mask = botmask, i=4; i<12; i++) {
			mask |= bitexr[ src[srcpitch*(i+8)-1] ];
			mask |=    - !! src[srcpitch*(i+8)  ];
			mask |= bitexl[ src[srcpitch*(i+8)+1] ];
			hmasks[i+4] |= mask;
		  }
		  // Check the 3x1.5 cells below
		  for(mask = botmask, i=11; i>=4; i--) {
			mask |= bitexr[ src[srcpitch*(i-8)-1] ];
			mask |=    - !! src[srcpitch*(i-8)  ];
			mask |= bitexl[ src[srcpitch*(i-8)+1] ];
			hmasks[i+4] |= mask;
		  }
		  for(mask = cenmask,i=7; i>=0; i--) {
			mask |= bitexr[ src[srcpitch*(i-8)-1] ];
			mask |=    - !! src[srcpitch*(i-8)  ];
			mask |= bitexl[ src[srcpitch*(i-8)+1] ];
			hmasks[i+4] |= mask;
		  }
		  for(mask = topmask, i=3; i>=-4; i--) {
			mask |= bitexr[ src[srcpitch*(i-8)-1] ];
			mask |=    - !! src[srcpitch*(i-8)  ];
			mask |= bitexl[ src[srcpitch*(i-8)+1] ];
			hmasks[i+4] |= mask;
		  }
		  // count the halo pixels
		  for(i=0; i<16; i++)
			alpha2 += bitcnt[hmasks[i]];
		}
		else {
		  // How many lit pixels in the centre cell?
		  for(i=0; i<8; i++)
			alpha1 += bitcnt[src[srcpitch*i]];
		  alpha1 *=2;

		  if (alpha1) {
			// If we have any lit pixels we fully occupy the cell.
			alpha2 = 128;
		  }
		  else {
			// No lit pixels here so build the halo mask from the neighbours
			BYTE cenmask = 0;

			// Check left and right neighbours, extend the halo
			// mask 8 pixels in from the nearest lit pixels.
			for(i=0; i<8; i++) {
			  cenmask |= bitexr[src[srcpitch*i-1]];
			  cenmask |= bitexl[src[srcpitch*i+1]];
			}

			if (cenmask == 0xFF) {
			  // If we have hard adjacent lit pixels we fully occupy this cell.
			  alpha2 = 128;
			}
			else {
			  BYTE hmasks[8], mask;

			  mask = cenmask;
			  for(i=0; i<8; i++) {
				// Check the 3 cells above
				mask |= bitexr[ src[srcpitch*(i+8)-1] ];
				mask |=    - !! src[srcpitch*(i+8)  ];
				mask |= bitexl[ src[srcpitch*(i+8)+1] ];
				hmasks[i] = mask;
			  }

			  mask = cenmask;
			  for(i=7; i>=0; i--) {
				// Check the 3 cells below
				mask |= bitexr[ src[srcpitch*(i-8)-1] ];
				mask |=    - !! src[srcpitch*(i-8)  ];
				mask |= bitexl[ src[srcpitch*(i-8)+1] ];

				alpha2 += bitcnt[hmasks[i] | mask];
			  }
			  alpha2 *=2;
			}
		  }
		}
		alpha2  = gamma[alpha2];
		alpha1  = gamma[alpha1];

		alpha2 -= alpha1;        
		alpha2 *= Ahalo;
		alpha1 *= Atext;
        // Pre calulate table for quick use  --  Pc = (Pc * dest[0] + dest[c]) >> 8;

		dest[0] = (64*516*255 - alpha1 -          alpha2)>>15;
		dest[1] = (    BVtext * alpha1 + BVhalo * alpha2)>>15;
		dest[2] = (    GUtext * alpha1 + GUhalo * alpha2)>>15;
		dest[3] = (    RYtext * alpha1 + RYhalo * alpha2)>>15;
      }
	  else {
		dest[0] = 256;
		dest[1] = 0;
		dest[2] = 0;
		dest[3] = 0;
      }

      dest += 4;
      ++src;
    } while(--wt);
  }

  xl=w-xl;
  xr=w-xr;
}






/************************************
 *******   Helper Functions    ******
 ***********************************/

void ApplyMessage( PVideoFrame* frame, const VideoInfo& vi, const char* message, int size, 
                   int textcolor, int halocolor, int bgcolor, IScriptEnvironment* env ) 
{
  if (vi.IsYUV()) {
    textcolor = RGB2YUV(textcolor);
    halocolor = RGB2YUV(halocolor);
  }
  Antialiaser antialiaser(vi.width, vi.height, "Arial", size, textcolor, halocolor);
  HDC hdcAntialias = antialiaser.GetDC();
  if  (hdcAntialias) {
	RECT r = { 4*8, 4*8, (vi.width-4)*8, (vi.height-4)*8 };
	DrawText(hdcAntialias, message, lstrlen(message), &r, DT_NOPREFIX|DT_CENTER);
	GdiFlush();
	antialiaser.Apply(vi, frame, (*frame)->GetPitch());
  }
}

